use std::cmp::min;
use std::collections::HashMap;
use std::error::Error;
use std::fs::{File, OpenOptions};
use std::io::{BufRead, BufReader, BufWriter, Write};
use std::path::Path;
use std::str::Bytes;
use std::time::Instant;

use vec_vp_tree::VpTree;

const PATH_QUERY: &str = "./query.txt";
const PATH_DATASET: &str = "./dataset.txt";
const PATH_RESULT: &str = "./result.txt";

#[derive(Clone)]
struct SimilarStringPtr<'a> {
    value: &'a str,
    distance: usize,
}

static mut GLOBAL_LEVENSHTEIN_COUNTER: u32 = 0;


const MODE_BRUTEFORCE: &'static str = "bruteforce";
const MODE_VANTAGE_POINT: &'static str = "vpt";

fn main() {
    let args: Vec<String> = std::env::args().collect();
    let mode = if args.len() != 2 {
        MODE_BRUTEFORCE
    } else {
        args[1].as_str()
    };

    let query = read_lines(PATH_QUERY, 100).unwrap();
    let dataset = read_lines(PATH_DATASET, 100_000).unwrap();
    let mut result: HashMap<&str, SimilarStringPtr> = HashMap::with_capacity(
        query.len() + query.len() / 10
    );

    let start = Instant::now();
    match mode {
        MODE_BRUTEFORCE => bruteforce(&query, &dataset, &mut result),
        MODE_VANTAGE_POINT => vpt(&query, &dataset, &mut result),
        mode => panic!("Unsupported mode: {}", mode),
    }

    let elapsed = start.elapsed();
    println!("Elapsed: {}", elapsed.as_millis());
    println!("Levenshteins: {}", unsafe { GLOBAL_LEVENSHTEIN_COUNTER });

    //save_result(PATH_RESULT, result).unwrap();
}

fn read_lines<P: AsRef<Path>>(path: P, estimated_capacity: usize) -> Result<Vec<String>, Box<dyn Error>> {
    let file = File::open(path)?;
    let mut reader = BufReader::with_capacity(1024 * 16, file);
    let mut data = Vec::with_capacity(estimated_capacity);
    let mut buffer = std::string::String::with_capacity(512);

    while reader.read_line(&mut buffer)? > 0 {
        let line = buffer.trim();
        data.push(line.to_owned());
        buffer.clear();
    }

    Ok(data)
}


fn save_result<P: AsRef<Path>>(destination: P, result: HashMap<&str, SimilarStringPtr>) -> Result<(), Box<dyn Error>> {
    let file = OpenOptions::new()
        .create(true)
        .write(true)
        .truncate(true)
        .open(destination)?;

    let mut writer = BufWriter::with_capacity(1024 * 8, file);
    for (k, v) in result {
        writeln!(writer, "{},{},{}", k, v.value, v.distance)?;
    }
    writer.flush()?;

    Ok(())
}


#[inline]
fn levenshtein_u8_str<S: AsRef<str>>(a: &S, b: &S, max: usize) -> usize {
    struct AsciiStrWrapper<'a> (&'a str);
    impl<'a, 'b: 'a, > IntoIterator for &'a AsciiStrWrapper<'b> {
        type Item = u8;
        type IntoIter = Bytes<'b>;

        fn into_iter(self) -> Self::IntoIter {
            self.0.bytes()
        }
    }
    unsafe {
        GLOBAL_LEVENSHTEIN_COUNTER += 1;
    }

    levenshtein(
        &AsciiStrWrapper(a.as_ref()),
        &AsciiStrWrapper(b.as_ref()),
        max,
    )
}

pub fn levenshtein<'a, 'b, Iter1, Iter2, Elem1, Elem2>(a: &'a Iter1, b: &'b Iter2, max_distance: usize) -> usize
    where &'a Iter1: IntoIterator<Item=Elem1>,
          &'b Iter2: IntoIterator<Item=Elem2>,
          Elem1: PartialEq<Elem2>
{
    let b_len = b.into_iter().count();
    if a.into_iter().next().is_none() {
        return b_len;
    }

    let mut cache: Vec<usize> = (1..b_len + 1).collect();
    let mut result = 0;

    for (i, a_elem) in a.into_iter().enumerate() {
        result = i + 1;
        let mut distance_b = i;
        let mut early_termination_distance = i;

        for (j, b_elem) in b.into_iter().enumerate() {
            let cost = if a_elem == b_elem {
                0usize
            } else {
                1usize
            };

            let distance_a = distance_b + cost;
            distance_b = cache[j];
            result = min(
                result + 1,
                min(distance_a, distance_b + 1),
            );
            cache[j] = result;

            if result < early_termination_distance {
                early_termination_distance = result;
            }
        }

        if early_termination_distance > max_distance {
            return max_distance;
        }
    }

    result
}


fn bruteforce<'a>(query: &'a [String], dataset: &'a [String], result: &mut HashMap<&'a str, SimilarStringPtr<'a>>) {
    for (row_num, q) in query.iter().enumerate()
    {
        let mut min_distance = usize::MAX;
        let mut nearest = None;

        let start = Instant::now();
        for entry in dataset.iter() {
            let distance = levenshtein_u8_str(q, entry, min_distance);

            if distance < min_distance {
                min_distance = distance;
                nearest = Some(entry);
            }
        }

        if let Some(entry) = nearest {
            result.insert(q.as_str(), SimilarStringPtr {
                value: entry.as_str(),
                distance: min_distance,
            });
        }

        println!("{:3}|> Elapsed: {:5}; Distance: {:3}", row_num, start.elapsed().as_millis(), min_distance);
    }
}

fn vpt<'a>(query: &'a [String], dataset: &'a [String], result: &mut HashMap<&'a str, SimilarStringPtr<'a>>) {
    let mut start = Instant::now();
    let tree = VpTree::new_with_dist(dataset, |a: &&String, b: &&String| levenshtein_u8_str(a, b, usize::MAX) as u64);
    let elapsed_tree_construction = start.elapsed().as_millis();
    let construction_levenshteins = unsafe { GLOBAL_LEVENSHTEIN_COUNTER };

    start = Instant::now();
    for (row_num, q) in query.iter().enumerate() {
        let query_start = Instant::now();
        let nearest = tree.k_nearest(q, 1);
        result.insert(q.as_str(), SimilarStringPtr {
            value: nearest[0].item.as_str(),
            distance: nearest[0].dist as usize,
        });
        println!("{:3}|> Elapsed: {:5}; Distance: {:3}", row_num, query_start.elapsed().as_millis(), nearest[0].dist);
    }

    println!("Tree constructed in: {:6} with {} levenshteins", elapsed_tree_construction, construction_levenshteins);
    println!("Searching took: {:6} with {} levenshteins", start.elapsed().as_millis(), unsafe { GLOBAL_LEVENSHTEIN_COUNTER - construction_levenshteins });
}


